@@ -1,21 +1,19 @@
 from os.path import abspath, exists, basename, dirname, join, realpath
 from os import makedirs, unlink, readlink, rmdir
-try:
-  from commands import getstatusoutput
-except ImportError:
-  from subprocess import getstatusoutput
+from alibuild_helpers import __version__
 from alibuild_helpers.analytics import report_event
 from alibuild_helpers.log import debug, error, info, banner, warning
 from alibuild_helpers.log import dieOnError
-from alibuild_helpers.cmd import execute, getStatusOutputBash, BASH
+from alibuild_helpers.cmd import execute, getstatusoutput, DockerRunner, BASH, install_wrapper_script
 from alibuild_helpers.utilities import star, prunePaths
-from alibuild_helpers.utilities import format, dockerStatusOutput, parseDefaults, readDefaults
+from alibuild_helpers.utilities import resolve_store_path
+from alibuild_helpers.utilities import format, parseDefaults, readDefaults
 from alibuild_helpers.utilities import getPackageList
 from alibuild_helpers.utilities import validateDefaults
 from alibuild_helpers.utilities import Hasher
 from alibuild_helpers.utilities import yamlDump
 from alibuild_helpers.utilities import resolve_tag, resolve_version
-from alibuild_helpers.git import partialCloneFilter
+from alibuild_helpers.git import git, partialCloneFilter
 from alibuild_helpers.sync import (NoRemoteSync, HttpRemoteSync, S3RemoteSync,
                                    Boto3RemoteSync, RsyncRemoteSync)
 import yaml
@@ -32,39 +30,27 @@ try:
 except ImportError:
   from pipes import quote  # Python 2.7
 
+import importlib
 import socket
 import os
-import ssl
 import re
 import shutil
 import sys
 import time
 
-def gzip():
-  return getstatusoutput("which pigz")[0] and "gzip" or "pigz"
-
-def tar():
-  return getstatusoutput("tar --ignore-failed-read -cvvf /dev/null /dev/zero")[0] and "tar" or "tar --ignore-failed-read"
 
 def writeAll(fn, txt):
   f = open(fn, "w")
   f.write(txt)
   f.close()
 
+
 def readHashFile(fn):
   try:
     return open(fn).read().strip("\n")
   except IOError:
     return "0"
 
-def getDirectoryHash(d):
-  if exists(join(d, ".git")):
-    err, out = getstatusoutput("GIT_DIR=%s/.git git rev-parse HEAD" % d)
-    dieOnError(err, "Impossible to find reference for %s " % d)
-  else:
-    err, out = getstatusoutput("pip --disable-pip-version-check show alibuild | grep -e \"^Version:\" | sed -e 's/.* //'")
-    dieOnError(err, "Impossible to find reference for %s " % d)
-  return out
 
 # Creates a directory in the store which contains symlinks to the package
 # and its direct / indirect dependencies
@@ -115,50 +101,124 @@ def storeHashes(package, specs, isDevelPkg, considerRelocation):
     # subsequent calculations.
     return
 
-  h = Hasher()
-  dh = Hasher()
-
-  for x in ["recipe", "version", "package", "commit_hash",
-            "env", "append_path", "prepend_path"]:
-    if sys.version_info[0] < 3 and x in spec and type(spec[x]) == OrderedDict:
-      # Python 2: use YAML dict order to prevent changing hashes
-      h(str(yaml.safe_load(yamlDump(spec[x]))))
-    else:
-      h(str(spec.get(x, "none")))
-
-  # If the commit hash is a real hash, and not a tag, we can safely assume
-  # that's unique, and therefore we can avoid putting the repository or the
-  # name of the branch in the hash.
-  if spec["commit_hash"] == spec.get("tag", "0"):
-    h(spec.get("source", "none"))
-    if "source" in spec:
-      h(spec["tag"])
-
-  for dep in spec.get("requires", []):
-    # At this point, our dependencies have a single hash, local or remote.
-    h(specs[dep]["hash"])
-    dh(specs[dep]["hash"] + specs[dep].get("devel_hash", ""))
+  # For now, all the hashers share data -- they'll be split below.
+  h_all = Hasher()
 
   if spec.get("force_rebuild", False):
-    h(str(time.time()))
+    h_all(str(time.time()))
+
+  def hash_data_for_key(key):
+    if sys.version_info[0] < 3 and key in spec and isinstance(spec[key], OrderedDict):
+      # Python 2: use YAML dict order to prevent changing hashes
+      return str(yaml.safe_load(yamlDump(spec[key])))
+    else:
+      return str(spec.get(key, "none"))
+
+  for x in ("recipe", "version", "package"):
+    h_all(hash_data_for_key(x))
+
+  # commit_hash could be a commit hash (if we're not building a tag, but
+  # instead e.g. a branch or particular commit specified by its hash), or it
+  # could be a tag name (if we're building a tag). We want to calculate the
+  # hash for both cases, so that if we build some commit, we want to be able to
+  # reuse tarballs from other builds of the same commit, even if it was
+  # referred to differently in the other build.
+  debug("Base git ref is %s", spec["commit_hash"])
+  h_default = h_all.copy()
+  h_default(spec["commit_hash"])
+  try:
+    # If spec["commit_hash"] is a tag, get the actual git commit hash.
+    real_commit_hash = spec["git_refs"]["refs/tags/" + spec["commit_hash"]]
+  except KeyError:
+    # If it's not a tag, assume it's an actual commit hash.
+    real_commit_hash = spec["commit_hash"]
+  # Get any other git tags that refer to the same commit. We do not consider
+  # branches, as their heads move, and that will cause problems.
+  debug("Real commit hash is %s, storing alternative", real_commit_hash)
+  h_real_commit = h_all.copy()
+  h_real_commit(real_commit_hash)
+  h_alternatives = [(spec.get("tag", "0"), spec["commit_hash"], h_default),
+                    (spec.get("tag", "0"), real_commit_hash, h_real_commit)]
+  for ref, git_hash in spec.get("git_refs", {}).items():
+    if ref.startswith("refs/tags/") and git_hash == real_commit_hash:
+      tag_name = ref[len("refs/tags/"):]
+      debug("Tag %s also points to %s, storing alternative",
+            tag_name, real_commit_hash)
+      hasher = h_all.copy()
+      hasher(tag_name)
+      h_alternatives.append((tag_name, git_hash, hasher))
+
+  # Now that we've split the hasher with the real commit hash off from the ones
+  # with a tag name, h_all has to add the data to all of them separately.
+  def h_all(data):  # pylint: disable=function-redefined
+    for _, _, hasher in h_alternatives:
+      hasher(data)
+
+  for x in ("env", "append_path", "prepend_path"):
+    h_all(hash_data_for_key(x))
+
+  for tag, commit_hash, hasher in h_alternatives:
+    # If the commit hash is a real hash, and not a tag, we can safely assume
+    # that's unique, and therefore we can avoid putting the repository or the
+    # name of the branch in the hash.
+    if commit_hash == tag:
+      hasher(spec.get("source", "none"))
+      if "source" in spec:
+        hasher(tag)
+
+  dh = Hasher()
+  for dep in spec.get("requires", []):
+    # At this point, our dependencies have a single hash, local or remote, in
+    # specs[dep]["hash"].
+    hash_and_devel_hash = specs[dep]["hash"] + specs[dep].get("devel_hash", "")
+    # If this package is a dev package, and it depends on another dev pkg, then
+    # this package's hash shouldn't change if the other dev package was
+    # changed, so that we can just rebuild this one incrementally.
+    h_all(specs[dep]["hash"] if isDevelPkg else hash_and_devel_hash)
+    # The deps_hash should always change, however, so we actually rebuild the
+    # dependent package (even if incrementally).
+    dh(hash_and_devel_hash)
 
   if isDevelPkg and "incremental_recipe" in spec:
-    h(spec["incremental_recipe"])
+    h_all(spec["incremental_recipe"])
     ih = Hasher()
     ih(spec["incremental_recipe"])
     spec["incremental_hash"] = ih.hexdigest()
   elif isDevelPkg:
-    h(spec.get("devel_hash"))
+    h_all(spec.get("devel_hash"))
 
   if considerRelocation and "relocate_paths" in spec:
-    h("relocate:"+" ".join(sorted(spec["relocate_paths"])))
+    h_all("relocate:"+" ".join(sorted(spec["relocate_paths"])))
 
-  spec["remote_revision_hash"] = h.hexdigest()
   spec["deps_hash"] = dh.hexdigest()
+  spec["remote_revision_hash"] = h_default.hexdigest()
+  # Store hypothetical hashes of this spec if we were building it using other
+  # tags that refer to the same commit that we're actually building. These are
+  # later used when fetching from the remote store. The "primary" hash should
+  # be the first in the list, so it's checked first by the remote stores.
+  spec["remote_hashes"] = [spec["remote_revision_hash"]] + \
+    list({h.hexdigest() for _, _, h in h_alternatives} - {spec["remote_revision_hash"]})
   # The local hash must differ from the remote hash to avoid conflicts where
   # the remote has a package with the same hash as an existing local revision.
-  h("local")
-  spec["local_revision_hash"] = h.hexdigest()
+  h_all("local")
+  spec["local_revision_hash"] = h_default.hexdigest()
+  spec["local_hashes"] = [spec["local_revision_hash"]] + \
+    list({h.hexdigest() for _, _, h, in h_alternatives} - {spec["local_revision_hash"]})
+
+
+def better_tarball(spec, old, new):
+  """Return which tarball we should prefer to reuse."""
+  if not old: return new
+  if not new: return old
+  old_rev, old_hash, _ = old
+  new_rev, new_hash, _ = new
+  old_is_local, new_is_local = old_rev.startswith("local"), new_rev.startswith("local")
+  # If one is local and one is remote, return the remote one.
+  if old_is_local and not new_is_local: return new
+  if new_is_local and not old_is_local: return old
+  # Finally, return the one that appears in the list of hashes earlier.
+  hashes = spec["local_hashes" if old_is_local else "remote_hashes"]
+  return old if hashes.index(old_hash) < hashes.index(new_hash) else new
 
 
 def doBuild(args, parser):
@@ -171,7 +231,7 @@ def doBuild(args, parser):
     syncHelper = Boto3RemoteSync(args.remoteStore, args.writeStore,
                                  args.architecture, args.workDir)
   elif args.remoteStore:
-    syncHelper = RsyncRemoteSync(args.remoteStore, args.writeStore, args.architecture, args.workDir, "")
+    syncHelper = RsyncRemoteSync(args.remoteStore, args.writeStore, args.architecture, args.workDir)
   else:
     syncHelper = NoRemoteSync()
 
@@ -183,11 +243,13 @@ def doBuild(args, parser):
   prunePaths(workDir)
 
   if not exists(args.configDir):
-    return (error, ("Cannot find %sdist recipes under directory \"%s\".\n" +
-                    "Maybe you need to \"cd\" to the right directory or " +
-                    "you forgot to run \"aliBuild init\"?") % (star(), args.configDir), 1)
+    error('Cannot find %sdist recipes under directory "%s".\n'
+          'Maybe you need to "cd" to the right directory or '
+          'you forgot to run "%sBuild init"?',
+          star(), args.configDir, star())
+    return 1
 
-  err, value = getstatusoutput("GIT_DIR=%s/.git git symbolic-ref -q HEAD" % args.configDir)
+  _, value = git(("symbolic-ref", "-q", "HEAD"), directory=args.configDir, check=False)
   branch_basename = re.sub("refs/heads/", "", value)
   branch_stream = re.sub("-patches$", "", branch_basename)
   # In case the basename and the stream are the same,
@@ -204,41 +266,48 @@ def doBuild(args, parser):
   if not exists(specDir):
     makedirs(specDir)
 
-  os.environ["ALIBUILD_ALIDIST_HASH"] = getDirectoryHash(args.configDir)
+  os.environ["ALIBUILD_ALIDIST_HASH"] = git(("rev-parse", "HEAD"), directory=args.configDir)
 
   debug("Building for architecture %s", args.architecture)
   debug("Number of parallel builds: %d", args.jobs)
   debug("Using %sBuild from %sbuild@%s recipes in %sdist@%s",
-        star(), star(), getDirectoryHash(dirname(__file__)), star(),
-        os.environ["ALIBUILD_ALIDIST_HASH"])
+        star(), star(), __version__, star(), os.environ["ALIBUILD_ALIDIST_HASH"])
+
+  install_wrapper_script("git", workDir)
+
+  with DockerRunner(dockerImage, ["--network=host"]) as getstatusoutput_docker:
+    my_gzip = "pigz" if getstatusoutput_docker("which pigz")[0] == 0 else "gzip"
+    my_tar = ("tar --ignore-failed-read"
+              if getstatusoutput_docker("tar --ignore-failed-read -cvvf "
+                                        "/dev/null /dev/zero")[0] == 0
+              else "tar")
+    systemPackages, ownPackages, failed, validDefaults = \
+      getPackageList(packages                = packages,
+                     specs                   = specs,
+                     configDir               = args.configDir,
+                     preferSystem            = args.preferSystem,
+                     noSystem                = args.noSystem,
+                     architecture            = args.architecture,
+                     disable                 = args.disable,
+                     force_rebuild           = args.force_rebuild,
+                     defaults                = args.defaults,
+                     performPreferCheck      = lambda pkg, cmd: getstatusoutput_docker(cmd),
+                     performRequirementCheck = lambda pkg, cmd: getstatusoutput_docker(cmd),
+                     performValidateDefaults = lambda spec: validateDefaults(spec, args.defaults),
+                     overrides               = overrides,
+                     taps                    = taps,
+                     log                     = debug)
 
-  (systemPackages, ownPackages, failed, validDefaults) = getPackageList(packages                = packages,
-                                                                        specs                   = specs,
-                                                                        configDir               = args.configDir,
-                                                                        preferSystem            = args.preferSystem,
-                                                                        noSystem                = args.noSystem,
-                                                                        architecture            = args.architecture,
-                                                                        disable                 = args.disable,
-                                                                        defaults                = args.defaults,
-                                                                        dieOnError              = dieOnError,
-                                                                        performPreferCheck      = lambda pkg, cmd : dockerStatusOutput(cmd, dockerImage, executor=getStatusOutputBash),
-                                                                        performRequirementCheck = lambda pkg, cmd : dockerStatusOutput(cmd, dockerImage, executor=getStatusOutputBash),
-                                                                        performValidateDefaults = lambda spec : validateDefaults(spec, args.defaults),
-                                                                        overrides               = overrides,
-                                                                        taps                    = taps,
-                                                                        log                     = debug)
   if validDefaults and args.defaults not in validDefaults:
-    return (error, "Specified default `%s' is not compatible with the packages you want to build.\n" % args.defaults +
-                   "Valid defaults:\n\n- " +
-                   "\n- ".join(sorted(validDefaults)), 1)
+    error("Specified default `%s' is not compatible with the packages you want to build.\n"
+          "Valid defaults:\n\n- %s", args.defaults, "\n- ".join(sorted(validDefaults)))
+    return 1
 
   if failed:
-    return (error,
-            "The following packages are system requirements and could not be found:\n\n- {failed}\n\n"
-            "Please run:\n\n\t{star}Doctor --defaults {defaults} {pkgname}\n\nto get a full diagnosis."
-            .format(star=star(), pkgname=args.pkgname.pop(), defaults=args.defaults,
-                    failed="\n- ".join(sorted(list(failed)))),
-            1)
+    error("The following packages are system requirements and could not be found:\n\n- %s\n\n"
+          "Please run:\n\n\t%sDoctor --defaults %s %s\n\nto get a full diagnosis.",
+          "\n- ".join(sorted(list(failed))), star(), args.defaults, args.pkgname.pop())
+    return 1
 
   for x in specs.values():
     x["requires"] = [r for r in x["requires"] if not r in args.disable]
@@ -279,16 +348,19 @@ def doBuild(args, parser):
               "hour": str(now.hour).zfill(2) }
 
   # Check if any of the packages can be picked up from a local checkout
-  develCandidates = [basename(d) for d in glob("*") if os.path.isdir(d)]
-  develCandidatesUpper = [basename(d).upper() for d in glob("*") if os.path.isdir(d)]
-  develPkgs = [p for p in buildOrder
-               if p in develCandidates and p not in args.noDevel]
-  develPkgsUpper = [(p, p.upper()) for p in buildOrder
-                    if p.upper() in develCandidatesUpper and p not in args.noDevel]
-  if set(develPkgs) != set(x for (x, y) in develPkgsUpper):
-    return (error, format("The following development packages have wrong spelling: %(pkgs)s.\n"
-                          "Please check your local checkout and adapt to the correct one indicated.",
-                          pkgs=", ".join(set(x.strip() for (x,y) in develPkgsUpper) - set(develPkgs))), 1)
+  develPkgs = []
+  if not args.forceTracked:
+    develCandidates = [basename(d) for d in glob("*") if os.path.isdir(d)]
+    develCandidatesUpper = [basename(d).upper() for d in glob("*") if os.path.isdir(d)]
+    develPkgs = [p for p in buildOrder
+                 if p in develCandidates and p not in args.noDevel]
+    develPkgsUpper = [(p, p.upper()) for p in buildOrder
+                      if p.upper() in develCandidatesUpper and p not in args.noDevel]
+    if set(develPkgs) != {x for x, _ in develPkgsUpper}:
+      error("The following development packages have the wrong spelling: %s.\n"
+            "Please check your local checkout and adapt to the correct one indicated.",
+            ", ".join({x.strip() for x, _ in develPkgsUpper} - set(develPkgs)))
+      return 1
 
   if buildOrder:
     banner("Packages will be built in the following order:\n - %s",
@@ -311,23 +383,25 @@ def doBuild(args, parser):
       updateReferenceRepoSpec(args.referenceSources, p, specs[p], args.fetchRepos, not args.docker)
 
       # Retrieve git heads
-      cmd = "git ls-remote --heads %s" % (specs[p].get("reference", specs[p]["source"]))
+      cmd = ("ls-remote", "--heads", "--tags",
+             specs[p].get("reference", specs[p]["source"]))
       if specs[p]["package"] in develPkgs:
          specs[p]["source"] = join(os.getcwd(), specs[p]["package"])
-         cmd = "git ls-remote --heads %s" % specs[p]["source"]
-      debug("Executing %s", cmd)
-      err, output = getStatusOutputBash(cmd)
-      if err:
-        raise RuntimeError("Error on '%s': %s" % (cmd, output))
-      specs[p]["git_heads"] = output.split("\n")
-      return "ok"
-    future_to_download = {executor.submit(downloadTask, p): p for p in [p for p in buildOrder if "source" in specs[p]]}
+         cmd = "ls-remote", "--heads", "--tags", specs[p]["source"]
+      output = git(cmd)
+      specs[p]["git_refs"] = {git_ref: git_hash for git_hash, sep, git_ref in
+                              (line.partition("\t") for line in output.splitlines())
+                              if sep}
+      return "%d refs found" % len(specs[p]["git_refs"])
+    future_to_download = {executor.submit(downloadTask, p): p
+                          for p in buildOrder if "source" in specs[p]}
     for future in concurrent.futures.as_completed(future_to_download):
         futurePackage = future_to_download[future]
         try:
             data = future.result()
         except Exception as exc:
-            raise RuntimeError("Error on fetching '%r'. Aborting." % futurePackage)
+            raise RuntimeError("Error on fetching %r: %s. Aborting." %
+                               (futurePackage, exc))
         else:
             debug("%r package updated: %s", futurePackage, data)
 
@@ -339,39 +413,33 @@ def doBuild(args, parser):
     if "source" in spec:
       # Tag may contain date params like %(year)s, %(month)s, %(day)s, %(hour).
       spec["tag"] = resolve_tag(spec)
-      # By default we assume tag is a commit hash. We then try to find
-      # out if the tag is actually a branch and we use the tip of the branch
-      # as commit_hash. Finally if the package is a development one, we use the
-      # name of the branch as commit_hash.
-      spec["commit_hash"] = spec["tag"]
-      for head in spec["git_heads"]:
-        if head.endswith("refs/heads/{0}".format(spec["tag"])) or spec["package"] in develPkgs:
-          spec["commit_hash"] = head.split("\t", 1)[0]
-          # We are in development mode, we need to rebuild if the commit hash
-          # is different and if there are extra changes on to.
-          if spec["package"] in develPkgs:
-            # Devel package: we get the commit hash from the checked source, not from remote.
-            cmd = "cd %s && git rev-parse HEAD" % spec["source"]
-            err, out = getstatusoutput(cmd)
-            dieOnError(err, "Unable to detect current commit hash.")
-            spec["commit_hash"] = out.strip()
-            cmd = "cd %s && git diff -r HEAD && git status --porcelain" % spec["source"]
-            h = Hasher()
-            err = execute(cmd, lambda s, *a: h(s % a))
-            debug("Command %s returned %d", cmd, err)
-            dieOnError(err, "Unable to detect source code changes.")
-            spec["devel_hash"] = spec["commit_hash"] + h.hexdigest()
-            cmd = "cd %s && git rev-parse --abbrev-ref HEAD" % spec["source"]
-            err, out = getstatusoutput(cmd)
-            if out == "HEAD":
-              err, out = getstatusoutput("cd %s && git rev-parse HEAD" % spec["source"])
-              out = out[0:10]
-            if err:
-              return (error, "Error, unable to lookup changes in development package %s. Is it a git clone?" % spec["source"], 1)
-            develPackageBranch = out.replace("/", "-")
-            spec["tag"] = args.develPrefix if "develPrefix" in args else develPackageBranch
-            spec["commit_hash"] = "0"
-          break
+      # First, we try to resolve the "tag" as a branch name, and use its tip as
+      # the commit_hash. If it's not a branch, it must be a tag or a raw commit
+      # hash, so we use it directly. Finally if the package is a development
+      # one, we use the name of the branch as commit_hash.
+      assert "git_refs" in spec
+      try:
+        spec["commit_hash"] = spec["git_refs"]["refs/heads/" + spec["tag"]]
+      except KeyError:
+        spec["commit_hash"] = spec["tag"]
+      # We are in development mode, we need to rebuild if the commit hash is
+      # different or if there are extra changes on top.
+      if spec["package"] in develPkgs:
+        # Devel package: we get the commit hash from the checked source, not from remote.
+        out = git(("rev-parse", "HEAD"), directory=spec["source"])
+        spec["commit_hash"] = out.strip()
+        cmd = "cd %s && git diff -r HEAD && git status --porcelain" % spec["source"]
+        h = Hasher()
+        err = execute(cmd, lambda s, *a: h(s % a))
+        debug("Command %s returned %d", cmd, err)
+        dieOnError(err, "Unable to detect source code changes.")
+        spec["devel_hash"] = spec["commit_hash"] + h.hexdigest()
+        out = git(("rev-parse", "--abbrev-ref", "HEAD"), directory=spec["source"])
+        if out == "HEAD":
+          out = git(("rev-parse", "HEAD"), directory=spec["source"])[:10]
+        develPackageBranch = out.replace("/", "-")
+        spec["tag"] = args.develPrefix if "develPrefix" in args else develPackageBranch
+        spec["commit_hash"] = "0"
 
     # Version may contain date params like tag, plus %(commit_hash)s,
     # %(short_hash)s and %(tag)s.
@@ -389,7 +457,8 @@ def doBuild(args, parser):
   # If one of the special packages is in the list of packages to be built,
   # we use it as main package, rather than the last one.
   if not buildOrder:
-    return (banner, "Nothing to be done.", 0)
+    banner("Nothing to be done.")
+    return 0
   mainPackage = buildOrder[-1]
   mainHash = specs[mainPackage]["commit_hash"]
 
@@ -431,9 +500,16 @@ def doBuild(args, parser):
     # a build_requires only anymore, so we drop it from the list.
     spec["full_build_requires"] = set(spec["full_build_requires"]) - spec["full_runtime_requires"]
 
+  # Use the selected plugin to build, instead of the default behaviour, if a
+  # plugin was selected.
+  if args.plugin != "legacy":
+    return importlib.import_module("alibuild_helpers.%s_plugin" % args.plugin) \
+                    .build_plugin(specs, args, buildOrder)
+
   debug("We will build packages in the following order: %s", " ".join(buildOrder))
   if args.dryRun:
-    return (info, "--dry-run / -n specified. Not building.", 0)
+    info("--dry-run / -n specified. Not building.")
+    return 0
 
   # We now iterate on all the packages, making sure we build correctly every
   # single one of them. This is done this way so that the second time we run we
@@ -453,7 +529,9 @@ def doBuild(args, parser):
   while buildOrder:
     packageIterations += 1
     if packageIterations > 20:
-      return (error, "Too many attempts at building %s. Something wrong with the repository?" % spec["package"], 1)
+      error("Too many attempts at building %s. Something wrong with the repository?",
+            spec["package"])
+      return 1
     p = buildOrder[0]
     spec = specs[p]
     if args.debug:
@@ -471,30 +549,8 @@ def doBuild(args, parser):
     debug("develPkgs = %r", develPkgs)
     storeHashes(p, specs, isDevelPkg=p in develPkgs,
                 considerRelocation=args.architecture.startswith("osx"))
-    debug("Hashes for recipe %s are %s (remote), %s (local)",
-          p, spec["remote_revision_hash"], spec["local_revision_hash"])
-
-    # This adds to the spec where it should find, locally or remotely the
-    # various tarballs and links.
-    pkgSpec = {
-      "workDir": workDir,
-      "package": spec["package"],
-      "version": spec["version"],
-      "remote_revision_hash": spec["remote_revision_hash"],
-      "remote_prefix": spec["remote_revision_hash"][0:2],
-      "local_revision_hash": spec["local_revision_hash"],
-      "local_prefix": spec["local_revision_hash"][0:2],
-      "architecture": args.architecture
-    }
-    spec.update({k: v % pkgSpec for k, v in (
-      ("remote_store_path", "TARS/%(architecture)s/store/%(remote_prefix)s/%(remote_revision_hash)s"),
-      ("remote_links_path", "TARS/%(architecture)s/%(package)s"),
-      ("remote_tar_hash_dir",
-       "%(workDir)s/TARS/%(architecture)s/store/%(remote_prefix)s/%(remote_revision_hash)s"),
-      ("local_tar_hash_dir",
-       "%(workDir)s/TARS/%(architecture)s/store/%(local_prefix)s/%(local_revision_hash)s"),
-      ("remote_tar_link_dir", "%(workDir)s/TARS/%(architecture)s/%(package)s"),
-    )})
+    debug("Hashes for recipe %s are %s (remote); %s (local)", p,
+          ", ".join(spec["remote_hashes"]), ", ".join(spec["local_hashes"]))
 
     if spec["package"] in develPkgs and getattr(syncHelper, "writeStore", None):
       warning("Disabling remote write store from now since %s is a development package.", spec["package"])
@@ -560,23 +616,25 @@ def doBuild(args, parser):
     if spec["package"] == mainPackage:
       mainBuildFamily = spec["build_family"]
 
+    candidate = None
     busyRevisions = set()
     # We can tell that the remote store is read-only if it has an empty or
     # no writeStore property. See below for explanation of why we need this.
     revisionPrefix = "" if getattr(syncHelper, "writeStore", "") else "local"
-    for d in packages:
-      realPath = readlink(d)
+    for symlink in packages:
+      realPath = readlink(symlink)
       matcher = format("../../%(a)s/store/[0-9a-f]{2}/([0-9a-f]*)/%(p)s-%(v)s-((?:local)?[0-9]*).%(a)s.tar.gz$",
                        a=args.architecture,
                        p=spec["package"],
                        v=spec["version"])
-      m = re.match(matcher, realPath)
-      if not m:
+      match = re.match(matcher, realPath)
+      if not match:
+        warning("Symlink %s -> %s couldn't be parsed", symlink, realPath)
         continue
-      h, revision = m.groups()
+      rev_hash, revision = match.groups()
 
-      if not (("local" in revision and h == spec["local_revision_hash"]) or
-              ("local" not in revision and h == spec["remote_revision_hash"])):
+      if not (("local" in revision and rev_hash in spec["local_hashes"]) or
+              ("local" not in revision and rev_hash in spec["remote_hashes"])):
         # This tarball's hash doesn't match what we need. Remember that its
         # revision number is taken, in case we assign our own later.
         if revision.startswith(revisionPrefix) and revision[len(revisionPrefix):].isdigit():
@@ -588,54 +646,51 @@ def doBuild(args, parser):
       # Don't re-use local revisions when we have a read-write store, so that
       # packages we'll upload later don't depend on local revisions.
       if getattr(syncHelper, "writeStore", False) and "local" in revision:
+        debug("Skipping revision %s because we want to upload later", revision)
         continue
 
       # If we have an hash match, we use the old revision for the package
-      # and we do not need to build it.
-      spec["revision"] = revision
-      if spec["package"] in develPkgs and "incremental_recipe" in spec:
-        spec["obsolete_tarball"] = d
-      else:
-        debug("Package %s with hash %s is already found in %s. Not building.", p, h, d)
-        src = format("%(v)s-%(r)s",
-                     w=workDir,
-                     v=spec["version"],
-                     r=spec["revision"])
-        dst1 = format("%(w)s/%(a)s/%(p)s/latest-%(bf)s",
-                      w=workDir,
-                      a=args.architecture,
-                      p=spec["package"],
-                      bf=spec["build_family"])
-        dst2 = format("%(w)s/%(a)s/%(p)s/latest",
-                      w=workDir,
-                      a=args.architecture,
-                      p=spec["package"])
+      # and we do not need to build it. Because we prefer reusing remote
+      # revisions, only store a local revision if there is no other candidate
+      # for reuse yet.
+      candidate = better_tarball(spec, candidate, (revision, rev_hash, symlink))
 
-        getstatusoutput("ln -snf %s %s" % (src, dst1))
-        getstatusoutput("ln -snf %s %s" % (src, dst2))
-        info("Using cached build for %s", p)
-      break
-
-    # If we aren't using an existing revision, assign the next free revision
-    # to this package. If we're not uploading it, name it localN to avoid
-    # interference with the remote store -- in case this package is built
-    # somewhere else, the next revision N might be assigned there, and would
-    # conflict with our revision N.
-    if "revision" not in spec:
+    try:
+      revision, rev_hash, symlink = candidate
+    except TypeError:  # raised if candidate is still None
+      # If we can't reuse an existing revision, assign the next free revision
+      # to this package. If we're not uploading it, name it localN to avoid
+      # interference with the remote store -- in case this package is built
+      # somewhere else, the next revision N might be assigned there, and would
+      # conflict with our revision N.
       # The code finding busyRevisions above already ensures that revision
       # numbers start with revisionPrefix, and has left us plain ints.
       spec["revision"] = revisionPrefix + str(
         min(set(range(1, max(busyRevisions) + 2)) - busyRevisions)
         if busyRevisions else 1)
+    else:
+      spec["revision"] = revision
+      # Remember what hash we're actually using.
+      spec["local_revision_hash" if revision.startswith("local")
+           else "remote_revision_hash"] = rev_hash
+      if spec["package"] in develPkgs and "incremental_recipe" in spec:
+        spec["obsolete_tarball"] = symlink
+      else:
+        debug("Package %s with hash %s is already found in %s. Not building.",
+              p, rev_hash, symlink)
+        getstatusoutput(
+          "ln -snf {v}-{r} {w}/{a}/{p}/latest-{bf};"
+          "ln -snf {v}-{r} {w}/{a}/{p}/latest".format(
+            v=spec["version"], r=spec["revision"], w=workDir,
+            a=args.architecture, p=spec["package"], bf=spec["build_family"]))
+        info("Using cached build for %s", p)
 
     # Now we know whether we're using a local or remote package, so we can set
     # the proper hash and tarball directory.
     if spec["revision"].startswith("local"):
       spec["hash"] = spec["local_revision_hash"]
-      spec["tar_hash_dir"] = spec["local_tar_hash_dir"]
     else:
       spec["hash"] = spec["remote_revision_hash"]
-      spec["tar_hash_dir"] = spec["remote_tar_hash_dir"]
     spec["old_devel_hash"] = readHashFile(join(
       workDir, "BUILD", spec["hash"], spec["package"], ".build_succeeded"))
 
@@ -753,13 +808,14 @@ def doBuild(args, parser):
     # directory contains files with non-ASCII names, e.g. Golang/Boost.
     shutil.rmtree(dirname(hashFile).encode("utf-8"), True)
 
-    debug("Looking for cached tarball in %s", spec["tar_hash_dir"])
+    tar_hash_dir = os.path.join(workDir, resolve_store_path(args.architecture, spec["hash"]))
+    debug("Looking for cached tarball in %s", tar_hash_dir)
     # FIXME: I should get the tar_hash_dir updated with server at this point.
     #        It does not really matter that the symlinks are ok at this point
     #        as I only used the tarballs as reusable binary blobs.
     spec["cachedTarball"] = ""
     if not spec["package"] in develPkgs:
-      tarballs = glob(join(spec["tar_hash_dir"], "*gz"))
+      tarballs = glob(os.path.join(tar_hash_dir, "*gz"))
       spec["cachedTarball"] = tarballs[0] if len(tarballs) else ""
       debug("Found tarball in %s" % spec["cachedTarball"]
             if spec["cachedTarball"] else "No cache tarballs found")
@@ -923,8 +979,8 @@ def doBuild(args, parser):
       ("GIT_COMMITTER_NAME", "unknown"),
       ("GIT_COMMITTER_EMAIL", "unknown"),
       ("GIT_TAG", spec["tag"]),
-      ("MY_GZIP", gzip()),
-      ("MY_TAR", tar()),
+      ("MY_GZIP", my_gzip),
+      ("MY_TAR", my_tar),
       ("INCREMENTAL_BUILD_HASH", spec.get("incremental_hash", "0")),
       ("JOBS", str(args.jobs)),
       ("PKGHASH", spec["hash"]),
@@ -945,13 +1001,12 @@ def doBuild(args, parser):
     # will perform the actual build. Otherwise build as usual using bash.
     if args.docker:
       dockerWrapper = (
-        "docker run --rm --entrypoint= --user $(id -u):$(id -g) "
+        "docker run --rm --network=host --entrypoint= --user $(id -u):$(id -g) "
         "-v {workdir}:/sw -v {scriptDir}/build.sh:/build.sh:ro "
         "-e GIT_REFERENCE_OVERRIDE=/mirror -e WORK_DIR_OVERRIDE=/sw "
         "{mirrorVolume} {develVolumes} {additionalEnv} {additionalVolumes} "
-        "{overrideSource} {extraArgs} {image} {bash} -ex /build.sh"
+        "{overrideSource} {extraArgs} {image} bash -ex /build.sh"
       ).format(
-        bash=quote(BASH),
         image=quote(dockerImage),
         workdir=quote(abspath(args.workDir)),
         scriptDir=quote(scriptDir),
@@ -1029,5 +1084,5 @@ def doBuild(args, parser):
   for x in develPkgs:
     banner("Build directory for devel package %s:\n%s/BUILD/%s-latest%s/%s",
            x, abspath(args.workDir), x, "-"+args.develPrefix if "develPrefix" in args else "", x)
-  return (debug, "Everything done", 0)
-
+  debug("Everything done")
+  return 0
