@@ -1,10 +1,6 @@
 #!/usr/bin/env python
-import subprocess, yaml
-try:
-  from commands import getstatusoutput
-except ImportError:
-  from subprocess import getstatusoutput
-from os.path import dirname, exists
+import yaml
+from os.path import exists
 import hashlib
 from glob import glob
 from os.path import basename, join
@@ -16,42 +12,38 @@ try:
   from collections import OrderedDict
 except ImportError:
   from ordereddict import OrderedDict
-try:
-  from shlex import quote  # Python 3.3+
-except ImportError:
-  from pipes import quote  # Python 2.7
+
+from alibuild_helpers.cmd import decode_with_fallback, getoutput
+from alibuild_helpers.git import git
+from alibuild_helpers.log import dieOnError
+
 
 class SpecError(Exception):
   pass
 
 asList = lambda x : x if type(x) == list else [x]
 
-# Keep the linter happy
-if sys.version_info[0] >= 3:
-  basestring = None
-  unicode = None
-
 def star():
   return re.sub("build.*$", "", basename(sys.argv[0]).lower())
 
-def is_string(s):
-  if sys.version_info[0] >= 3:
-    return isinstance(s, str)
-  return isinstance(s, basestring)
 
-def to_unicode(s):
-  if sys.version_info[0] >= 3:
-    if isinstance(s, bytes):
-      try:
-        return s.decode("utf-8")  # to get newlines as such and not as escaped \n
-      except:
-        return s.decode("latin-1")  # Workaround issue with some special characters of latin-1 which are not understood by unicode
-    return str(s)
-  elif isinstance(s, str):
-    return unicode(s, "utf-8")  # utf-8 is a safe assumption
-  elif not isinstance(s, unicode):
-    return unicode(str(s))
-  return s
+def resolve_store_path(architecture, spec_hash):
+  """Return the path where a tarball with the given hash is to be stored.
+
+  The returned path is relative to the working directory (normally sw/) or the
+  root of the remote store.
+  """
+  return "/".join(("TARS", architecture, "store", spec_hash[:2], spec_hash))
+
+
+def resolve_links_path(architecture, package):
+  """Return the path where symlinks for the given package are to be stored.
+
+  The returned path is relative to the working directory (normally sw/) or the
+  root of the remote store.
+  """
+  return "/".join(("TARS", architecture, package))
+
 
 # Date fields to substitute: they are zero-padded
 now = datetime.now()
@@ -133,8 +125,10 @@ def validateDefaults(finalPkgSpec, defaults):
                    defaults,
                    "\n".join([" - " + x for x in validDefaults])), validDefaults)
 
+
 def format(s, **kwds):
-  return to_unicode(s) % kwds
+  return decode_with_fallback(s) % kwds
+
 
 def doDetectArch(hasOsRelease, osReleaseLines, platformTuple, platformSystem, platformProcessor):
   if platformSystem == "Darwin":
@@ -173,11 +167,7 @@ def doDetectArch(hasOsRelease, osReleaseLines, platformTuple, platformSystem, pl
   processor = platformProcessor
   if not processor:
     # Sometimes platform.processor returns an empty string
-    stdout, _ = subprocess.Popen(
-      ("uname", "-m"), stdout=subprocess.PIPE,
-      stderr=None, stdin=None
-    ).communicate()
-    processor = stdout.decode("ascii").strip()
+    processor = getoutput(("uname", "-m")).strip()
 
   return "{distro}{version}_{machine}".format(
     distro=distribution, version=version.split(".")[0],
@@ -225,6 +215,12 @@ def filterByArchitecture(arch, requires):
     if re.match(matcher, arch):
       yield require
 
+def disabledByArchitecture(arch, requires):
+  for r in requires:
+    require, matcher = ":" in r and r.split(":", 1) or (r, ".*")
+    if not re.match(matcher, arch):
+      yield require
+
 def readDefaults(configDir, defaults, error, architecture):
   defaultsFilename = "%s/defaults-%s.sh" % (configDir, defaults)
   if not exists(defaultsFilename):
@@ -270,9 +266,9 @@ class GitReader(object):
     self.url, self.configDir = url, configDir
   def __call__(self):
     m = re.search(r'^dist:(.*)@([^@]+)$', self.url)
-    fn,gh = m.groups()
-    err,d = getstatusoutput(format("GIT_DIR=%(dist)s/.git git show %(gh)s:%(fn)s.sh",
-                                   dist=self.configDir, gh=gh, fn=fn.lower()))
+    fn, gh = m.groups()
+    err, d = git(("show", "{gh}:{fn}.sh".format(gh=gh, fn=fn.lower())),
+                 directory=self.configDir)
     if err:
       raise RuntimeError(format("Cannot read recipe %(fn)s from reference %(gh)s.\n" +
                                 "Make sure you run first (this will not alter your recipes):\n" +
@@ -351,8 +347,8 @@ def parseDefaults(disable, defaultsGetter, log):
   return (None, overrides, taps)
 
 def getPackageList(packages, specs, configDir, preferSystem, noSystem,
-                   architecture, disable, defaults, dieOnError, performPreferCheck, performRequirementCheck,
-                   performValidateDefaults, overrides, taps, log):
+                   architecture, disable, defaults, performPreferCheck, performRequirementCheck,
+                   performValidateDefaults, overrides, taps, log, force_rebuild=()):
   systemPackages = set()
   ownPackages = set()
   failedRequirements = set()
@@ -417,6 +413,7 @@ def getPackageList(packages, specs, configDir, preferSystem, noSystem,
       else:
         disable.append(spec["package"])
 
+    spec["disabled"] = list(disable)
     if spec["package"] in disable:
       continue
 
@@ -429,6 +426,9 @@ def getPackageList(packages, specs, configDir, preferSystem, noSystem,
           validDefaults = None  # no valid default works for all current packages
 
     # For the moment we treat build_requires just as requires.
+    fn = lambda what: disabledByArchitecture(architecture, spec.get(what, []))
+    spec["disabled"] += [x for x in fn("requires")]
+    spec["disabled"] += [x for x in fn("build_requires")]
     fn = lambda what: filterByArchitecture(architecture, spec.get(what, []))
     spec["requires"] = [x for x in fn("requires") if not x in disable]
     spec["build_requires"] = [x for x in fn("build_requires") if not x in disable]
@@ -442,14 +442,11 @@ def getPackageList(packages, specs, configDir, preferSystem, noSystem,
     spec["tag"] = spec.get("tag", spec["version"])
     spec["version"] = spec["version"].replace("/", "_")
     spec["recipe"] = recipe.strip("\n")
+    spec["force_rebuild"] = spec["package"] in force_rebuild
     specs[spec["package"]] = spec
     packages += spec["requires"]
   return (systemPackages, ownPackages, failedRequirements, validDefaults)
 
-def dockerStatusOutput(cmd, dockerImage=None, executor=getstatusoutput):
-  return executor("docker run --rm --entrypoint= {image} bash -c {command}"
-                  .format(image=dockerImage, command=quote(cmd))
-                  if dockerImage else cmd)
 
 class Hasher:
   def __init__(self):
@@ -460,3 +457,7 @@ class Hasher:
     self.h.update(txt)
   def hexdigest(self):
     return self.h.hexdigest()
+  def copy(self):
+    new_hasher = Hasher()
+    new_hasher.h = self.h.copy()
+    return new_hasher
